use anyhow::Result;
use std::{
    any::{Any, TypeId},
    time::Duration,
};

use crate::{resources::Resource, world::ComponentChange, *};

enum CommandTypes {
    InsertEntity(EntityObject),
    RemoveEntity(Entity),
    UpdateEntity(Entity, ComponentChange),

    Event(TypeId, Box<dyn Any + Send + Sync>),

    InsertResource(TypeId, Box<dyn Resource>),
}

pub struct Commands {
    commands: Vec<CommandTypes>,
    now: Duration,
}

impl Commands {
    pub(crate) fn new(now: Duration) -> Self {
        Self {
            commands: Vec::new(),
            now,
        }
    }

    pub fn now(&self) -> Duration {
        self.now
    }

    pub fn insert_entity(&mut self, entity_object: EntityObject) -> Entity {
        let entity = entity_object.as_entity();

        self.commands
            .push(CommandTypes::InsertEntity(entity_object));

        entity
    }

    pub fn remove_entity(&mut self, entity: Entity) {
        self.commands.push(CommandTypes::RemoveEntity(entity));
    }

    pub fn insert_component<T: EntityComponent>(&mut self, entity: Entity, component: T) {
        self.commands.push(CommandTypes::UpdateEntity(
            entity,
            ComponentChange::Added(TypeId::of::<T>(), Box::new(component)),
        ));
    }

    pub fn remove_component<T: EntityComponent>(&mut self, entity: Entity) {
        self.commands.push(CommandTypes::UpdateEntity(
            entity,
            ComponentChange::Removed(TypeId::of::<T>()),
        ));
    }

    pub fn write_event<T: Any + Send + Sync>(&mut self, payload: T) {
        self.commands
            .push(CommandTypes::Event(TypeId::of::<T>(), Box::new(payload)));
    }

    pub fn insert_resource<T: Resource>(&mut self, resource: T) {
        self.commands.push(CommandTypes::InsertResource(
            TypeId::of::<T>(),
            Box::new(resource),
        ));
    }

    pub(crate) fn apply_deferred(self, world: &mut World) -> Result<()> {
        for command in self.commands {
            match command {
                CommandTypes::InsertEntity(entity_object) => {
                    world.add_entity(entity_object)?;
                }
                CommandTypes::RemoveEntity(entity) => world.remove_entity(entity),
                CommandTypes::UpdateEntity(entity, component_change) => {
                    world.component_change(entity, component_change)
                }
                CommandTypes::Event(type_id, payload) => {
                    world.events.write_payload(type_id, payload)
                }
                CommandTypes::InsertResource(type_id, resource) => {
                    world.resources.insert_resource(type_id, resource)
                }
            }
        }

        Ok(())
    }
}
